#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define VPCALC_SECTION	"DEV"
#define VPCALC_SUBSECT	"85"

#include "../config.h"
#include "../include/types.h"
#include "../include/debug.h"

#include "../include/mem.h"
#include "../include/gfx.h"

#include "../include/device.h"

#include "linkport.h"

typedef struct _device_o {
	uint8 keys[SDLK_LAST];
	uint8 keymask;
	uint8 key_sets[8];

	uint hold_onbut;

	uint8 onbutff;
	uint8 onstate, powermode;
	uint8 page;

	int active;

	int refreshing;
	int *low_power;

	mem_o mem;
	gfx_o gfx;
	linkport_o linkport;
} _device_o;

static int keymap_init = 0;
static uint16 keymap[SDLK_LAST];

static inline void _init_keymap(void)
{
	if (keymap_init)
		return;

	memset(keymap, 0, sizeof(keymap));

	keymap[SDLK_DOWN] = 0x0101;
	keymap[SDLK_LEFT] = 0x0102;
	keymap[SDLK_RIGHT] = 0x0104;
	keymap[SDLK_UP] = 0x0108;

	keymap[SDLK_RETURN] = 0x0201;
	keymap[SDLK_PLUS] = 0x0202;
	keymap[SDLK_MINUS] = 0x0204;
	keymap[SDLK_m] = 0x0208;
	keymap[SDLK_d] = 0x0210;
	keymap[SDLK_h] = 0x0220;
	keymap[SDLK_BACKSPACE] = 0x0240;

	keymap[SDLK_n] = 0x0301;
	keymap[SDLK_3] = 0x0302;
	keymap[SDLK_6] = 0x0304;
	keymap[SDLK_9] = 0x0308;
	keymap[SDLK_p] = 0x0310;
	keymap[SDLK_t] = 0x0320;
	keymap[SDLK_w] = 0x0340;

	keymap[SDLK_PERIOD] = 0x0401;
	keymap[SDLK_2] = 0x0402;
	keymap[SDLK_5] = 0x0404;
	keymap[SDLK_8] = 0x0408;
	keymap[SDLK_o] = 0x0410;
	keymap[SDLK_c] = 0x0420;
	keymap[SDLK_F8] = 0x0440;
	keymap[SDLK_DELETE] = 0x0480;

	keymap[SDLK_0] = 0x0501;
	keymap[SDLK_1] = 0x0502;
	keymap[SDLK_4] = 0x0504;
	keymap[SDLK_7] = 0x0508;
	keymap[SDLK_e] = 0x0510;
	keymap[SDLK_s] = 0x0520;
	keymap[SDLK_F7] = 0x0540;
	keymap[SDLK_x] = 0x0580;

	keymap[SDLK_LESS] = 0x0602;
	keymap[SDLK_COMMA] = 0x0604;
	keymap[SDLK_q] = 0x0608;
	keymap[SDLK_l] = 0x0610;
	keymap[SDLK_k] = 0x0620;
	keymap[SDLK_F6] = 0x0640;
	keymap[SDLK_LCTRL] = 0x0680;

	keymap[SDLK_F5] = 0x0701;
	keymap[SDLK_F4] = 0x0702;
	keymap[SDLK_F3] = 0x0704;
	keymap[SDLK_F2] = 0x0708;
	keymap[SDLK_F1] = 0x0710;
	keymap[SDLK_LSHIFT] = 0x0720;
	keymap[SDLK_ESCAPE] = 0x0740;
	keymap[SDLK_TAB] = 0x0780;

	keymap_init = 1;
}

device_o device_new(mem_o mem, gfx_o gfx)
{
	_device_o *new;
	int i;

	_init_keymap();
	new = (_device_o *) malloc(sizeof(_device_o));

	if (new != NULL) {
		memset(new, 0, sizeof(_device_o));

		for (i=0; i<8; i++)
			new->key_sets[i] = 0xff;

		new->keymask = 0xff;
		new->onstate = 0x00;
		new->hold_onbut = 10000;

		new->active = 0;
		new->refreshing = 0;
		new->low_power = NULL;

		new->mem = mem;
		new->page = 0x01;
		mem_remap(new->mem, 0, 0, PERM_READ | PERM_EXEC);
		mem_remap(new->mem, 1, 1, PERM_READ | PERM_EXEC);
		mem_remap(new->mem, 2, 0x10,
			PERM_READ | PERM_WRITE | PERM_EXEC);
		mem_remap(new->mem, 3, 0x11,
			PERM_READ | PERM_WRITE | PERM_EXEC);

		new->gfx = gfx;
		gfx_activate(new->gfx, 0);
		gfx_set_contrast(new->gfx, 0x0a);
		gfx_set_screen_addr(new->gfx, 0xc000);
		gfx_map_refresh(new->gfx, &(new->refreshing));

		new->linkport = linkport_new("dead");
		if (new->linkport == NULL) {
			free(new);
			return NULL;
		}
		linkport_update(new->linkport, 0x03);
	}

	return (device_o)new;
}

void device_del(device_o device)
{
	_device_o *_device = (_device_o *)device;

	linkport_del(_device->linkport);
	free(_device);
}

void device_keydown(device_o device, SDLKey key)
{
	_device_o *_device = (_device_o *)device;
	uint16 mapped_key = keymap[key];

	if ((key == SDLK_HASH) && (_device->low_power != NULL)) {
		*(_device->low_power) = 0;
	}

	_device->keys[key] = 0xff;

	_report("key down: %i (%04x)", key, mapped_key);

	_device->key_sets[mapped_key >> 8] &= ~(mapped_key & 0x00ff);
}

void device_keyup(device_o device, SDLKey key)
{
	_device_o *_device = (_device_o *)device;
	uint16 mapped_key = keymap[key];

	_device->keys[key] = 0;

	_report("key up: %i (%04x)", key, mapped_key);

	_device->key_sets[mapped_key >> 8] |= (mapped_key & 0x00ff);
}

void device_out(device_o device, uint port, uint data)
{
	_device_o *_device = (_device_o *)device;
	uint16 scr_addr;

	switch(port & 0xff)
	{
	case 0x00:		// LCD Screen Pointer
		scr_addr = ((data & 0x3f) + 0xc0) << 8;
		gfx_set_screen_addr(_device->gfx, scr_addr);
		return;

	case 0x01:		// Keypad Interface
		_device->keymask = data & 0xff;
		return;

	case 0x02:		// Contrast Control
		data &= 0x1f;

		if(data & 0x10)
			data = - (data | 0xf0);

		gfx_set_contrast(_device->gfx, ((data & 0x0f) + 1) << 1);
		return;

	case 0x03:		// onstate
		if ((data ^ _device->onstate) & 0x02) {
			_device->active = (data & 0x02);
			gfx_activate(_device->gfx, data & 0x02);
		}

		if ((_device->low_power != NULL) &&
			((~data ^ *(_device->low_power)) & 0x08) &&
			!_device->hold_onbut) {
			*(_device->low_power) = ~data & 0x08;
		}

		_device->onstate = data & 0x0f;
		return;

	case 0x04:		// ???
		return;

	case 0x05:		// Swappable Mem page
        	if(_device->page != data) {
			_device->page = data;

			data &= 0x7;

			mem_remap(_device->mem, 1, data,
				PERM_READ | PERM_EXEC);
		}
		return;

	case 0x06:		// powermode
		_device->powermode = data;
		return;

	case 0x07:		// Link Port
		linkport_update(_device->linkport, data & 0x3c);
		return;

	default:
		_warn_report("unknown output port! %i", port & 0xff);
		return;
	}
}

uint device_in(device_o device, uint port)
{
	_device_o *_device = (_device_o *)device;
	uint8 btmp, state;

	switch(port & 0xFF)
	{
	case 0x03:		// onstate
		if (_device->hold_onbut) {
			_device->hold_onbut--;

			return 0x01;
		}

		state = _device->keys[SDLK_HASH];

		btmp = 0;

		if (_device->refreshing)
			btmp |= 0x04;
		_device->refreshing ^= 0x04;

		if (_device->active)
			btmp |= 0x02;

		if(_device->onbutff != state) {
			_device->onbutff = state;

			if(state)
				btmp |= 0x01;
		}

		if(!state)
			return (btmp | 0x08);

		return (btmp & 0xf7);

	case 0x05:		// Swappable memory page
		return _device->page;

	case 0x06:		// powerstate
		return _device->powermode;

 	case 0x07:		// link
		return linkport_status(_device->linkport);

	case 0x01:		// keypad
		state = _device->keymask;
		btmp = 0xff;

		if (!(state & 0x01))
			btmp &= _device->key_sets[1];
		if (!(state & 0x02))
			btmp &= _device->key_sets[2];
		if (!(state & 0x04))
			btmp &= _device->key_sets[3];
		if (!(state & 0x08))
			btmp &= _device->key_sets[4];
		if (!(state & 0x10))
			btmp &= _device->key_sets[5];
		if (!(state & 0x20))
			btmp &= _device->key_sets[6];
		if (!(state & 0x40))
			btmp &= _device->key_sets[7];
		return btmp;

	default:
		_warn_report("unknown input port! %i", port & 0xff);
		return port;
  	}
}

void device_map_sleep(device_o device, int *ptr)
{
	_device_o *_device = (_device_o *)device;

	_device->low_power = ptr;
}
