#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>

#define VPCALC_SECTION	"LIB"
#define VPCALC_SUBSECT	"DEV"

#include "../config.h"
#include "../include/types.h"
#include "../include/debug.h"

#include "../include/mem.h"
#include "../include/gfx.h"

#include "../include/dev.h"
#include "../include/device.h"

typedef struct _dev_o {
	void *library;

	device_o device;

	void (*del)(device_o device);

	void (*keydown)(device_o dev, SDLKey key);
	void (*keyup)(device_o dev, SDLKey key);

	void (*out)(device_o dev, uint port, uint data);
	uint (*in)(device_o dev, uint port);

	void (*map_sleep)(device_o dev, int *ptr);
} _dev_o;

dev_o dev_new(const char *libname, mem_o mem, gfx_o gfx)
{
	_dev_o *new;
	device_o (*device_new)(mem_o, gfx_o);
	const char *error;
  
	new = (_dev_o *) malloc(sizeof(_dev_o));
	if (new == NULL) {
		_err_report("Unable to allocate cpu structure.");

		return NULL;
	}

	new->library = dlopen(libname, RTLD_NOW);
	if (!new->library) {
		error = dlerror();
		_err_report(error);
		free(new);
		return NULL;
	}

	device_new = dlsym(new->library, "device_new");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->del = dlsym(new->library, "device_del");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->keydown = dlsym(new->library, "device_keydown");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->keyup = dlsym(new->library, "device_keyup");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->out = dlsym(new->library, "device_out");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->in = dlsym(new->library, "device_in");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->map_sleep = dlsym(new->library, "device_map_sleep");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->device = device_new(mem, gfx);
	if (new->device == NULL) {
		_err_report("failed to make new device unit.");

		dlclose(new->library);
		free(new);
		return NULL;
	}

	return (dev_o)new;
}

void dev_del(dev_o dev)
{
	_dev_o *_dev = (_dev_o *)dev;

	_dev->del(_dev->device);

	dlclose(_dev->library);
	free(_dev);
}

void dev_keydown(dev_o dev, SDLKey key)
{
	_dev_o *_dev = (_dev_o *)dev;

	_dev->keydown(_dev->device, key);
}

void dev_keyup(dev_o dev, SDLKey key)
{
	_dev_o *_dev = (_dev_o *)dev;

	_dev->keyup(_dev->device, key);
}

void dev_out(dev_o dev, uint port, uint data)
{
	_dev_o *_dev = (_dev_o *)dev;

	_dev->out(_dev->device, port, data);
}


uint dev_in(dev_o dev, uint port)
{
	_dev_o *_dev = (_dev_o *)dev;

	return _dev->in(_dev->device, port);

}

void dev_map_sleep(dev_o dev, int *ptr)
{
	_dev_o *_dev = (_dev_o *)dev;

	_dev->map_sleep(_dev->device, ptr);

}
